from dataclasses import dataclass


@dataclass
class OobaModelInfo:
    model_name: str | None
    lora_names: list[str]
    shared_settings: dict
    shared_args: dict


class OobaModelNotLoaded(OobaModelInfo):
    def __init__(self, **kwargs) -> None:
        super().__init__(model_name=None, lora_names=[], **kwargs)

    def __repr__(self) -> str:
        return "OobaModelNotLoaded()"



from typing import TYPE_CHECKING

import pydantic

from ooba_api.pydantic_compat import USING_PYDANTIC_LEGACY

if not TYPE_CHECKING:
    GuidanceScale = pydantic.confloat(ge=0, le=2.5)
    NewMaxTokens = pydantic.conint(gt=1)
    MinLength = pydantic.conint(gt=0)
    RepetitionPenaltyRange = pydantic.conint(gt=0)
    RepetitionPenalty = pydantic.confloat(gt=0)
    if USING_PYDANTIC_LEGACY:
        StoppingStrings = pydantic.conlist(str, min_items=0)
    else:
        StoppingStrings = pydantic.conlist(str, min_length=0)
    Temperature = pydantic.confloat(gt=0, lt=1.0)
    TopK = pydantic.conint(gt=0)
    TopP = pydantic.confloat(gt=0, lt=1)
    TypicalP = pydantic.conint(gt=0, lt=1)
    TruncationLength = pydantic.conint(gt=1)
else:
    GuidanceScale = float
    NewMaxTokens = int
    MinLength = int
    RepetitionPenaltyRange = int
    RepetitionPenalty = float
    StoppingStrings = list
    Temperature = float
    TopK = int
    TopP = float
    TypicalP = int
    TruncationLength = int


class Parameters(pydantic.BaseModel):
    # add beginning token, highly recommended but maybe some models don't want this
    add_bos_token: bool = True

    # ignore max_new_tokens and instead use math to set max tokens and max out the context length
    auto_max_new_tokens: bool = False

    # disables the "end of sentence" token. Probably would only enable this if combined with custom stopping strings.
    ban_eos_token: bool = False

    # knob, how much it will listen to you
    guidance_scale: GuidanceScale = 1.0

    # limit output tokens
    max_new_tokens: NewMaxTokens = 128

    # force generation of a minimum length
    min_length: MinLength = 0

    # knob, set the range, in tokens, to consider when applying the repetition penalty. 0 means disabled
    # a value like 4 helps prevent repeated junk (like newline comments)
    # a larger value or disabled helps prevent repeating itself, but may influence the output
    repetition_penalty_range: RepetitionPenaltyRange = 0

    # knob, penalize repeating tokens. Recommended
    # note that this may interfere with correct syntax in code generation.
    # Consider "foo(bar(1, 2))" becomes "foo(bar(1, 2)"
    repetition_penalty: RepetitionPenalty = 1.1

    # knob, random seed, use -1 for a random... random seed
    seed: int = -1

    # do not include special tokens. Some models have to disable this
    skip_special_tokens: bool = True

    # tell model to stop if it spits out one of these strings
    stopping_strings: StoppingStrings = []

    # knob, with sample, random selection. High is more random or creative
    temperature: Temperature = 0.6

    # knob, number of samples.
    top_k: TopK = 20

    # knob, probability threshold for considering a sample
    top_p: TopP = 0.5

    # knob, a value of 1 disables it. Tokens too far from this are removed.
    # Typically used if a high temperature starts producing gibberish.
    typical_p: TypicalP = 1

    # length to truncate the prompt. Should always be lower than the max context size
    truncation_length: TruncationLength = 4096


DEFAULT_PARAMETERS = Parameters()



import json
import logging
from multiprocessing import Lock

import requests

from ooba_api.model_info import OobaModelInfo, OobaModelNotLoaded
from ooba_api.parameters import DEFAULT_PARAMETERS, Parameters
from ooba_api.prompts import Prompt

logger = logging.getLogger("ooba_api")
prompt_logger = logging.getLogger("ooba_api.prompt")

# global lock
_one_at_a_time_lock = Lock()


class OobaApiClient:
    """
    Client for the Ooba Booga text generation web UI
    """

    # full URL to chat endpoint
    _chat_url: str

    # full URL to generate endpoint
    _generate_url: str

    # full URL to model endpoint
    _model_url: str

    # API Key, not yet used
    api_key: str | None

    # Enforce one request at a time to avoid overwhelming the server
    one_at_a_time: bool

    def __init__(
        self,
        url: str | None = None,
        *,
        host: str = "http://localhost",
        port: int = 5000,
        api_key: str | None = None,
        one_at_a_time: bool = True,
    ):
        if url:
            self.url = url
        else:
            self.url = f"{host}:{port}"
        self._chat_url = f"{self.url}/api/v1/chat"
        self._generate_url = f"{self.url}/api/v1/generate"
        self._model_url = f"{self.url}/api/v1/model"
        self.api_key = api_key
        self.one_at_a_time = one_at_a_time

        if self.api_key:
            logger.warning("API keys are not yet supported")

    def _post(self, target_url: str, timeout: float, data: dict) -> requests.Response:
        with _one_at_a_time_lock:
            return requests.post(target_url, timeout=timeout, json=data)

    def instruct(
        self,
        prompt: Prompt,
        parameters: Parameters = DEFAULT_PARAMETERS,
        timeout: int | float = 500,
        print_prompt: bool = False,
    ) -> str:
        """
        Provide an instruction, get a response

        :param messages: Message to provide an instruction
        :param max_tokens: Maximum tokens to generate
        :param timeout: When to timeout
        :param print_prompt: Print the prompt being used. Use case is debugging
        """
        prompt_to_use = prompt.full_prompt()
        if print_prompt:
            print(prompt_to_use)
        prompt_logger.info(prompt_to_use)

        # pydantic compatibility. dict -> model_dump
        if hasattr(parameters, "model_dump"):
            param_dict = parameters.model_dump()
        else:
            param_dict = parameters.dict()

        response = self._post(
            self._generate_url,
            timeout=timeout,
            data=(
                {"prompt": prompt_to_use, "negative_prompt": prompt.negative_prompt or ""}
                | param_dict
            ),
        )
        response.raise_for_status()
        data = response.json()
        if __debug__:
            logger.debug(json.dumps(data, indent=2))

        return data["results"][0]["text"]

    def _model_api(self, request: dict, timeout: int | float = 500) -> dict:
        response = self._post(self._model_url, timeout, request)
        response.raise_for_status()
        data = response.json()
        if __debug__:
            logger.debug(json.dumps(data, indent=2))

        return data["result"]

    def model_info(self) -> OobaModelInfo:
        result = self._model_api({"action": "info"}, timeout=5)
        model_name = result["model_name"]
        if model_name == "None":
            return OobaModelNotLoaded(
                shared_settings=result["shared.settings"], shared_args=result["shared.args"]
            )
        return OobaModelInfo(
            model_name=result["model_name"],
            lora_names=result["lora_names"],
            shared_settings=result["shared.settings"],
            shared_args=result["shared.args"],
        )

    def load_model(self, model_name: str, *, args_dict: dict) -> OobaModelInfo:
        result = self._model_api({"action": "load", "model_name": model_name, "args": args_dict}, timeout=5000)
        return OobaModelInfo(
            model_name=result["model_name"],
            lora_names=result["lora_names"],
            shared_settings=result["shared.settings"],
            shared_args=result["shared.args"],
        )



import textwrap
from typing import TYPE_CHECKING

import pydantic

from ooba_api.pydantic_compat import USING_PYDANTIC_LEGACY


class Prompt(pydantic.BaseModel):
    negative_prompt: str | None = None
    prompt: str

    def full_prompt(self) -> str:
        return self.prompt


class InstructPrompt(Prompt):
    """
    Used for instructions
    """

    instruct_template: str = "{prompt}"

    def full_prompt(self) -> str:
        return self.instruct_template.format(prompt=self.prompt)


if not TYPE_CHECKING:
    if USING_PYDANTIC_LEGACY:
        Messages = pydantic.conlist(dict, min_items=1)
    else:
        Messages = pydantic.conlist(dict, min_length=1)
else:
    Messages = list


class ChatPrompt(Prompt):
    """
    Used for chat

    Not tested, not implemented, just placeholder
    """

    messages: Messages


class LlamaInstructPrompt(Prompt):
    """
    Used for llama, llama 2, code llama, etc
    """

    system_prompt: str = ""
    instruct_template: str = textwrap.dedent(
        """
        [INST] <<SYS>> {system_prompt} <</SYS>> {user_prompt} [/INST]
        """
    ).strip()

    def full_prompt(self) -> str:
        return self.instruct_template.format(
            system_prompt=self.system_prompt, user_prompt=self.prompt
        )



from .clients import OobaApiClient
from .model_info import OobaModelInfo, OobaModelNotLoaded
from .parameters import Parameters
from .prompts import ChatPrompt, InstructPrompt, LlamaInstructPrompt, Prompt

__all__ = [
    "ChatPrompt",
    "InstructPrompt",
    "LlamaInstructPrompt",
    "OobaApiClient",
    "OobaModelInfo",
    "OobaModelNotLoaded",
    "Parameters",
    "Prompt",
]
